// Copyright (C) 2006 Erik Dahlberg
// 
// This file is part of LSystem3D.
// 
// LSystem3D is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// LSystem3D is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with LSystem3D; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA




#include <iostream>
#include <string>

#include <glibmm/stringutils.h>

#include "lsystemparameters.h"
#include "rule.h"
#include "ruleset.h"

using namespace std;



/**
 * Constructor
 */
LSystemParameters::LSystemParameters()
{
}



/**
 * Destructor
 */
LSystemParameters::~LSystemParameters()
{
}



/**
 * Load parameters from xml file
 * @param lsystem put parameters into this L-system
 * @param path path to xml file
 */
void LSystemParameters::load(LindenmayerSystem *lsystem, string path)
{
    if (lsystem)
    {
        try
        {
            // load
            loadFromDisk(path);
    
            // axiom
            findChild("axiom");
            Rule axiom("axiom", getString(), 1.0);
            lsystem->setAxiom(axiom);
            
            // rules
            if (findChild("rule"))
            {
                do
                {
                    // construct the rule
                    string name = getAttribute("name");
                    string content = getString();
                    double probability = Glib::Ascii::strtod(getAttribute("probability"));
                    Rule completeRule(name, content, probability);
                    
                    lsystem->setRule(completeRule);
                } while (findNextChild());
            }
    
            // angle
            findChild("angle");
            lsystem->setAngle(getNumber());
            
            // depth        
            findChild("depth");
            lsystem->setDepth((int)getNumber());
        }
        catch (xmlpp::exception e)
        {
            cerr << "LSystemParameters::load(): " << e.what() << endl;
        }
    }
    else
    {
        cerr << "invalid L-system" << endl;
    }
}



/**
 * Save parameters to xml file
 * @param lsystem get parameters from this L-system
 * @param path path to xml file
 */
void LSystemParameters::save(LindenmayerSystem *lsystem, string path)
{
    if (lsystem &&
        lsystem->getAxiom().getProbability() != 0.0 &&
        !lsystem->getRules().getRules().empty())
    {
        // new document with root node
        createDocumentWithRoot("lsystem");
        
        // axiom
        addChildToRoot("axiom");
        addString(lsystem->getAxiom().getContent());
        
        // rules
        rulemap rules = lsystem->getRules().getRules();
        for (rulemap::iterator it = rules.begin(); it != rules.end(); ++it)
        {
            addChildToRoot("rule");
            addString(it->second.getContent());
            
            addAttribute("name", it->second.getName());
            
            string probabilityString = Glib::Ascii::dtostr(it->second.getProbability());
            addAttribute("probability", probabilityString);
        }
        
        // angle
        addChildToRoot("angle");
        addNumber(lsystem->getAngle());
        
        // depth
        addChildToRoot("depth");
        addNumber(lsystem->getDepth());
        
        // save
        saveToDisk(path);
    }
    else
    {
        cerr << "invalid L-system" << endl;
    }
}
